
; zero-page
table_entry = $02
table_pos   = $03
volume      = $05
vbl_flag    = $06

; MIDI note range that can be played
first_note  = 43
last_note   = 84

ORG   $C000

reset:
      sei
      
      lda   #$40              ; disable APU frame IRQ
      sta   $4017
      lda   $4015             ; clear any already-pending IRQ
      
      lda   #$FF              ; DMC start addr = $FFC0
      sta   $4012
      lda   #0                ; DMC len = 1 sample
      sta   $4013
      lda   #$8D              ; initial values for IRQ handler
      sta   $4010
      lda   #0
      sta   table_pos
      
      lda   #0                ; init PPU and enable VBL
      sta   vbl_flag
      sta   $2001
      lda   #$80
      sta   $2000
      
replay:
      ldy   #first_note       ; play scale
next_note:
      tya
      jsr   saw_note
      
      ldx   #18               ; fade note out
decay:lda   vbl_flag          ; wait for VBL
      beq   decay
      dec   vbl_flag
      dec   volume
      dec   volume
      dex
      bne   decay
      iny                     ; next note
      cpy   #last_note+1
      bcc   next_note
      
      jmp   replay
      
; start saw note. a = MIDI note
saw_note:
      sec                     ; convert MIDI note to table index
      sbc   #first_note
      tax
      lda   note_table,x      ; set new table entry
      sta   table_entry
      
      lda   #48
      sta   volume
      
      cli                     ; start DMC if not already running
      lda   #$10
      sta   $4015
      rts


; NMI handler
nmi:
      inc   vbl_flag
      rti


; CPU cycles used per wave approx. = segment_count * 65 + 13
; (segment_count depends on current note)
      
                              ; # clocks for efficiency calculation
                              ; 7 for interrupt
irq:  pha                     ; 3
      txa                     ; 2
      pha                     ; 3
      
      ldx   table_pos         ; 3 next period entry
      lda   period_table,x    ; 4
      bne   mid               ; 3
                              ; -1
      lda   volume            ; 3 beginning of new saw
      sta   $4011             ; 4
      ldx   table_entry       ; 3
      lda   period_table,x    ; 4
mid:  sta   $4010             ; 4 set DMC period
      inx                     ; 2
      stx   table_pos         ; 4
      
      pla                     ; 4
      tax                     ; 2
      
      lda   #$10              ; 2 restart DMC as late as possible
      sta   $4015             ; 4
      
      pla                     ; 4
      rti                     ; 6

; DMC periods to use for each wave segment of a given note
period_table:
      dc.b 0
      dc.b 128, 128, 128, 128, 128, 138, 0
      dc.b 128, 128, 128, 130, 130, 136, 0
      dc.b 128, 128, 129, 132, 132, 134, 0
      dc.b 128, 128, 128, 129, 139, 139, 0
      dc.b 128, 128, 128, 134, 137, 138, 0
      dc.b 128, 128, 128, 131, 140, 0
      dc.b 128, 128, 128, 137, 141, 141, 0
      dc.b 128, 128, 129, 137, 139, 0
      dc.b 128, 128, 130, 141, 141, 142, 0
      dc.b 128, 128, 134, 136, 141, 0
      dc.b 128, 128, 130, 141, 0
      dc.b 128, 130, 135, 138, 141, 0
      dc.b 128, 128, 139, 141, 142, 0
      dc.b 128, 134, 137, 140, 141, 142, 0
      dc.b 128, 133, 140, 141, 142, 142, 0
      dc.b 128, 131, 140, 140, 0
      dc.b 128, 136, 137, 139, 0
      dc.b 128, 135, 139, 141, 0
      dc.b 129, 135, 139, 141, 0
      dc.b 128, 136, 142, 142, 0
      dc.b 129, 133, 141, 0
      dc.b 129, 135, 141, 0
      dc.b 128, 139, 141, 0
      dc.b 131, 139, 141, 142, 0
      dc.b 129, 140, 141, 0
      dc.b 133, 139, 141, 142, 0
      dc.b 136, 136, 139, 0
      dc.b 135, 137, 140, 0
      dc.b 134, 138, 141, 0
      dc.b 135, 138, 142, 0
      dc.b 134, 140, 142, 0
      dc.b 136, 140, 141, 0
      dc.b 136, 141, 141, 0
      dc.b 139, 140, 140, 0
      dc.b 138, 140, 142, 0
      dc.b 141, 142, 142, 142, 0
      dc.b 139, 141, 142, 0
      dc.b 139, 142, 142, 0
      dc.b 141, 141, 141, 0
      dc.b 141, 141, 142, 0
      dc.b 141, 142, 142, 0
      dc.b 142, 142, 142, 0


; offsets for notes
note_table:
      dc.b 1,8,15,22,29,36,42,49,55,62,68,73,79,85,92,99
      dc.b 104,109,114,119,124,128,132,136,141,145,150,154,158,162,166,170
      dc.b 174,178,182,186,191,195,199,203,207,211

ORG   $FFC0

      dc.b  0                 ; DMC sample
      
ORG   $FFFA

      dc.w  nmi
      dc.w  reset
      dc.w  irq
